"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum
from thbase.util.login import LoginEntry

TransportType = Enum('TransportType', ['FRAMED', 'BUFFERED'])
ProtocolType = Enum('ProtocolType', ['BINARY', 'COMPACT'])

HOST_DEFAULT = 'localhost'
PORT_DEFAULT = 9090
TRANSPORT_DEFAULT = TransportType.FRAMED
PROTOCOL_DEFAULT = ProtocolType.BINARY
RETRY_TIMEOUT_DEFAULT = 1
RETRY_TIMES_DEFAULT = 10
BATCH_SIZE_DEFAULT = 10
USE_SSL_DEFAULT = False
USE_HTTP_DEFAULT = False
RECONNECTION_TIMES = 10
RECONNECTION_TIMEOUT = 10


class ClientConfig(object):
    def __init__(self, thrift_host=HOST_DEFAULT,  # type: str
                 port=PORT_DEFAULT,  # type: int
                 retry_timeout=RETRY_TIMES_DEFAULT,  # type: int
                 retry_times=RETRY_TIMES_DEFAULT,  # type: int
                 connection_retry_times=RECONNECTION_TIMES,  # type: int
                 connection_retry_timeout=RECONNECTION_TIMEOUT,  # type: int
                 transport_type=TRANSPORT_DEFAULT,  # type: TransportType
                 protocol_type=PROTOCOL_DEFAULT,  # type: ProtocolType
                 use_ssl=USE_SSL_DEFAULT,  # type: bool
                 batch_size=BATCH_SIZE_DEFAULT,  # type: int
                 use_http=USE_HTTP_DEFAULT,  # type: bool
                 authentication=None
                 ):
        """
        Basic client configuration.
        Args:
            thrift_host: thrift server address
            port: thrift server port
            retry_timeout: time interval between two retries for operation. Unit is seconds.
            retry_times: operation retry times. If the operation is still unsuccessful after being
            retried for this times, it will return False.
            connection_retry_times: time interval between two retries for reconnection. Unit is seconds.
            connection_retry_timeout: reconnection retry times. If the client still cannot rebuild the connection
            after retrying for this times, an exception will be raised.
            transport_type: Up to now, the server support "framed", "buffered" and "http" transport.
            protocol_type: Up to now, the server support "Binary" and "Compact" protocol.
            use_ssl: if the client use SSL to secure the connection. Only support http transport and binary protocol.
            batch_size: the max size of the batch when using batch operation in a Table object.
            Batch size can be customized with given setter method.
            use_http: if the client use http as the transport but not TCP.
        """
        self._host = thrift_host
        self._port = port
        self._connection_retry_timeout = connection_retry_timeout
        self._connection_retry_times = connection_retry_times
        self._retry_timeout = retry_timeout
        self._retry_times = retry_times
        self._transport_type = transport_type
        self._protocol_type = protocol_type
        self._batch_size = batch_size
        self._use_ssl = use_ssl
        self._use_http = use_http
        self._authentication = authentication
        self._parameter_check()

    def _parameter_check(self):
        if not isinstance(self.port, int) or self.port not in range(65536):
            raise ValueError("Port must be an integer in 0~65535.")
        if not isinstance(self.retry_times, int) or self.retry_times < 1:
            raise ValueError("Retry times for connection rebuild must be a positive integer.")
        if not isinstance(self.retry_timeout, int) or self.retry_timeout < 1:
            raise ValueError("Time interval for connection rebuild must be a positive integer.")
        if not isinstance(self.batch_size, int) or self.batch_size < 1:
            raise ValueError("Batch size must be a positive integer.")
        if not isinstance(self.use_ssl, bool):
            raise ValueError("parameter use_ssl must be a bool value.")
        if not isinstance(self.use_http, bool):
            raise ValueError("parameter use_http must be a bool value.")
        if not (isinstance(self._authentication, LoginEntry) or None):
            raise ValueError("parameter authentication must be a LoginEntry object or None.")
        if self._transport_type not in TransportType:
            raise ValueError("Invalid type of transport {}. Use one of the specific enum type {}."
                             .format(type(self._transport_type), ', '.join([str(a) for a in TransportType])))
        if self._protocol_type not in ProtocolType:
            raise ValueError("Invalid type of protocol {}. Use one of the specific enum type {}."
                             .format(type(self._protocol_type), ', '.join([str(a) for a in ProtocolType])))

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def connection_retry_timeout(self):
        return self._connection_retry_timeout

    @property
    def connection_retry_times(self):
        return self._connection_retry_times

    @property
    def transport_type(self):
        return self._transport_type

    @property
    def protocol_type(self):
        return self._protocol_type

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def use_ssl(self):
        return self._use_ssl

    @property
    def use_http(self):
        return self._use_http

    @property
    def retry_times(self):
        return self._retry_times

    @property
    def retry_timeout(self):
        return self._retry_timeout

    @property
    def authentication(self):
        return self._authentication



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import time

from thrift.protocol import TBinaryProtocol, TCompactProtocol
from thrift.transport.TSocket import TSocket
from thrift.transport.TTransport import TBufferedTransport, TFramedTransport, TTransportException, TSaslClientTransport
from thrift.transport.THttpClient import THttpClient

from thbase.config import TransportType, ProtocolType

logger = logging.getLogger(__name__)

THRIFT_TRANSPORTS = {
    TransportType.BUFFERED: TBufferedTransport,
    TransportType.FRAMED: TFramedTransport,
}
THRIFT_PROTOCOLS = {
    ProtocolType.BINARY: TBinaryProtocol.TBinaryProtocol,
    ProtocolType.COMPACT: TCompactProtocol.TCompactProtocol,
}


class Connection(object):
    """
    Class to manage basic transport and protocol.
    User should not use instances of this class directly.
    The instances should be managed by a Client object.

    """
    def __init__(self, host,
                 port,
                 transport_type,
                 protocol_type,
                 retry_timeout,
                 retry_times,
                 use_ssl,
                 use_http,
                 authentication):

        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self.use_http = use_http
        self.authentication = authentication

        self._transport_type = THRIFT_TRANSPORTS[transport_type]
        self._protocol_type = THRIFT_PROTOCOLS[protocol_type]
        self._retry_timeout = retry_timeout
        self._retry_times = retry_times
        self._rebuild_protocol()
        self._initialized = True

    def _rebuild_protocol(self):
        """
        Rebuild the transport, protocol from the configuration.
        Should not be used directly by users.
        Returns:
            None
        """
        if self.use_http:
            # if use http transport,
            prefix = 'https://' if self.use_ssl else 'http://'
            self.transport = THttpClient(uri_or_host=prefix + self.host + ':' + str(self.port))
            self.protocol = TBinaryProtocol.TBinaryProtocol(self.transport)
            return

        if self.use_ssl:
            from thrift.transport.TSSLSocket import TSSLSocket
            socket = TSSLSocket(host=self.host, port=self.port, validate=False)
        else:
            socket = TSocket(host=self.host, port=self.port)

        if self.authentication:
            socket = TSaslClientTransport(socket, host=self.host,
                                          service=self.authentication.service,
                                          mechanism=self.authentication.mechanism,
                                          username=self.authentication.username,
                                          password=self.authentication.password)

        self.transport = self._transport_type(socket)
        self.protocol = self._protocol_type(self.transport)

    def is_open(self):
        return self.transport.isOpen()

    def open(self):
        if self.transport.isOpen():
            return
        logger.debug("Opening thrift transport throught TCP connection.")
        self.transport.open()

    def close(self):
        if not self.transport.isOpen():
            return
        if logger is not None:
            logger.debug("Closing thrift transport to {}:{}.".format(self.host, self.port))
        self.transport.close()

    def _reconnect(self):
        """
        Method to rebuild the connection with thrift server. Should not be used by the user directly.
        Returns: None

        """
        if not self.transport.isOpen():
            logger.info("Connection lose is detected and start reconnecting to the target thrift server.")
            for i in range(self._retry_times):
                if self.transport.isOpen():
                    logger.info("Reconnection success after retrying {} times.".format(i))
                    return True
                self._rebuild_protocol()
                try:
                    logger.info("Starting reconnection to thrift server.")
                    self.transport.open()
                    logger.info("Reconnection success after retrying {} times.".format(i + 1))
                    return True
                except TTransportException:
                    logger.error("Reconnected {} times but failed.".format(i + 1))
                    time.sleep(self._retry_timeout)
            if not self.transport.isOpen():
                logger.error("Failed to rebuild connection with target thrift server.")
                raise TTransportException(type=TTransportException.NOT_OPEN,
                                          message="Failed to rebuild connection with target thrift server.")
            return False



"""
Copyright 2021 Yutong Sean

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Union
from thbase.util.bytes import to_str


class Cell(object):

    def __init__(self, table_name,  # type: Union[None, bytes]
                 row,  # type: Union[None, bytes]
                 family,  # type: Union[None, bytes]
                 qualifier,  # type: Union[None, bytes]
                 value,  # type: Union[None, bytes]
                 timestamp,  # type: Union[None, int]
                 ):
        """
        Data structure to load data from hbase. If there is no matched cell or some errors occur at hbase server,
        all the attributes could be None. In python2, bytes are represented by str, so the type of value is str.
        Args:
            table_name: name of the table.
            row: the row key.
            family: the column family.
            qualifier: the column qualifier.
            value: the bytes stored in the cell.
            timestamp: a long int.
        """
        self._table_name = table_name
        self._row = row
        self._family = family
        self._qualifier = qualifier
        self._value = value
        self._timestamp = timestamp

    @property
    def table_name(self):
        return self._table_name

    @property
    def row(self):
        return self._row

    @property
    def family(self):
        return self._family

    @property
    def qualifier(self):
        return self._qualifier

    @property
    def value(self):
        return self._value

    @property
    def timestamp(self):
        return self._timestamp

    def __str__(self):
        return ":".join([to_str(self.table_name), to_str(self.row), to_str(self.family), to_str(self.qualifier)]) + \
               ' => ' + to_str(self.value)



from setuptools import setup, find_packages
from os import path
from sys import version_info
here = path.abspath(path.dirname(__file__))
if version_info.major == 3:
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    with open(path.join(here, 'README.md')) as f:
        long_description = f.read()

setup(
    name='thbase',
    version='1.3.7',
    description='HBase thrift2 client python API. Compatible with python2.7 and python3',
    long_description=long_description,
    long_description_content_type='text/markdown',
    # Author details
    author='Yutong Sean',
    author_email='yutongsean@gmail.com',
    url='https://github.com/YutSean',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3"
    ],
    packages=find_packages(),
    py_modules=["thbase"],
    install_requires=["thrift==0.13.0", "enum34", "typing", "pure-sasl"]
)
